<?php
/**
 * Font Stylesheet Inlining
 *
 * @package   Google\WordPress\Font_Stylesheet_Inlining
 * @license   GPL-2.0+
 * @copyright 2019 Google LLC
 *
 * @wordpress-plugin
 * Plugin Name: Font Stylesheet Inlining
 * Description: Improve frontend performance of external (Google) fonts by fetching provider stylesheets and inlining them. This should fix the Lighthouse audit "Ensure text remains visible during webfont load" and improve "Eliminate render-blocking resources". This eliminates a render blocking critical resource and <a href="https://web.dev/fast/avoid-invisible-text">avoids invisible text during font loading</a>. The font URL is automatically amended <code>display=fallback</code>, where the value this can be overridden via <code>wp_add_style_data( $handle, 'font-display', $value )</code> (see also <a href="https://core.trac.wordpress.org/ticket/47282">#47282</a>). When using with the <a href="https://wordpress.org/plugins/amp/">AMP plugin</a>, the size of the CSS in <code>style[amp-custom]</code> will increase by ~2KB, so beware of the 50KB budget. Also, beware that the fetched stylesheet will be requested with a Chrome user agent; it will not vary based on the current user agent accessing WordPress. Because of this, this must only be used for fonts which have suitable fallbacks in place for browsers that don't support the latest webfont format (woff2).
 * Plugin URI: https://gist.github.com/westonruter/25474fcdea0dc4ac69f3af551f0f15e9
 * Author: MeanThemes, Weston Ruter, Google
 * Author URI: https://www.meanthemes.com/
 * Version: 1.3
 * License: GPLv2 or later
 */

namespace Google\WordPress\Font_Stylesheet_Inlining;

/**
 * User agent for requesting font stylesheets.
 *
 * Note that font providers will vary the response based on the user agent, so using a user agent for the majority
 * of visitors will ensure that the inlined @font-face will work for the majority of visitors. Since no user-agent
 * varying is performed, it is important that only non-critical fonts be inlined. In other words, if a font has a suitable
 * fallback then it would be suitable for inlining since other non-supporting user agents would end up using the fallback
 * instead.
 *
 * @var string
 */
const USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Safari/537.36';

/**
 * Filter style tag to inline an external font stylesheet.
 *
 * Currently this only applies to Google Fonts.
 *
 * @param string $tag    The link tag for the enqueued style.
 * @param string $handle The style's registered handle.
 * @param string $href   The stylesheet's source URL.
 * @return string HTML link or style tag.
 */
function filter_style_tag( $tag, $handle, $href = '' ) {
	if ( ! preg_match( '#^(https?:)?//fonts\.googleapis\.com/css#', $href ) ) {
		return $tag;
	}

	if ( substr( $href, 0, 2 ) === '//' ) {
		$href = 'https:' . $href;
	}

	$result = fetch_font_stylesheet( $href, $handle );
	if ( is_wp_error( $result ) ) {
		$comment = __( 'Unable to inline font stylesheet.', 'font-stylesheet-inlining' ) . ' ' . $result->get_error_message();
		return sprintf( '<!-- %s -->', str_replace( '--', '—', esc_html( $comment ) ) ) . $tag;
	} else {
		if ( isset( wp_styles()->registered[ $handle ]->args ) ) {
			$media = wp_styles()->registered[ $handle ]->args;
		} else {
			$media = 'all';
		}
		return sprintf(
			'<style id="%s" media="%s" data-src="%s">%s</style>',
			esc_attr( "$handle-css" ),
			esc_attr( $media ),
			esc_url( $href ),
			wp_strip_all_tags( $result ) // Note that esc_html() cannot be used because `div &gt; span` is not interpreted properly.
		);
	}
}
add_filter( 'style_loader_tag', __NAMESPACE__ . '\filter_style_tag', 10, 3 );

/**
 * Fetch font stylesheet with font-display injected into @font-face rules.
 *
 * @param string $url    Font stylesheet URL.
 * @param string $handle Stylesheet handle.
 * @return string|\WP_Error Inlined stylesheet or error.
 */
function fetch_font_stylesheet( $url, $handle ) {
	if ( ! wp_style_is( $handle, 'registered' ) ) {
		return new \WP_Error( 'unregistered_style' );
	}

	// Replace '&#038;' with '&'.
	$url = html_entity_decode( $url, ENT_QUOTES, get_bloginfo( 'charset' ) );

	// Inject the font display param if not already present.
	if ( false === strpos( $url, '&display=' ) ) {
		$font_display = wp_styles()->get_data( $handle, 'font-display' );
		if ( ! $font_display ) {
			$font_display = 'swap';
		}
		$url = add_query_arg( 'display', $font_display, $url );
	}

	$transient_key = sprintf( 'font-stylesheet-%s-%s', $handle, md5( $url ) );

	$result = get_transient( $transient_key );
	if ( ! $result ) {
		$r = wp_remote_get(
			$url,
			array(
				'user-agent' => USER_AGENT,
				'headers'    => array(
					sprintf( 'Referer: %s', home_url( '/' ) ),
				),
			)
		);
		if ( 200 !== wp_remote_retrieve_response_code( $r ) ) {
			$result = new \WP_Error(
				'response_not_ok',
				sprintf(
					/* translators: %1$s is error message, %2$d is response code. */
					__( 'Unable to fetch stylesheet: %1$s (response code %2$d)', 'font-stylesheet-inlining' ),
					wp_remote_retrieve_response_message( $r ),
					wp_remote_retrieve_response_code( $r )
				)
			);
		} elseif ( 0 !== strpos( wp_remote_retrieve_header( $r, 'content-type' ), 'text/css' ) ) {
			$result = new \WP_Error(
				'response_not_css',
				sprintf(
					/* translators: %s is the content type returned. */
					__( 'Expected response to be text/css but got %s.', 'font-stylesheet-inlining' ),
					wp_remote_retrieve_header( $r, 'content-type' )
				)
			);
		} else {
			$result = wp_remote_retrieve_body( $r );
			if ( ! preg_match( '/@font-face\s*{/', $result ) ) {
				$result = new \WP_Error(
					'stylesheet_lacks_font_face',
					__( 'Stylesheet lacks @font-face rule.', 'font-stylesheet-inlining' )
				);
			}
		}

		$expiration = 0;
		if ( is_wp_error( $result ) ) {
			$expiration = \HOUR_IN_SECONDS;
		} elseif ( preg_match( '/max-age=(\d+)/', wp_remote_retrieve_header( $r, 'cache-control' ), $matches ) ) {
			$expiration = intval( $matches[1] );
		}
		set_transient( $transient_key, $result, max( \DAY_IN_SECONDS, $expiration ) );
	}

	if ( $result instanceof \WP_Error ) {
		return $result;
	}

	return $result;
}

/**
 * Filter resource hints to remove unnecessary DNS prefetch.
 *
 * @param array $urls The list of hints URLs.
 * @return array The list of hints URLs without fonts.googleapis.com.
 */
function filter_resource_hints( $urls ) {
	$return = array();
	foreach ( $urls as $url ) {
		if ( 'fonts.googleapis.com' !== $url ) {
			$return[] = $url;
		}
	}

	return $return;
}
add_filter( 'wp_resource_hints', __NAMESPACE__ . '\filter_resource_hints', 10 );
